"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.cli = void 0;
const utils_1 = require("./utils");
const constants_1 = require("./constants");
const evmAtomicTx_1 = require("./evmAtomicTx");
const pvmAtomicTx_1 = require("./pvmAtomicTx");
const addValidator_1 = require("./addValidator");
const dist_1 = require("@flarenetwork/flarejs/dist");
function cli(program) {
    return __awaiter(this, void 0, void 0, function* () {
        const validatorstring = { validator: program.STRING };
        program
            // global configurations
            .option("--network", "Network name (flare or costwo)", {
            validator: program.STRING,
            global: true,
            default: 'flare'
        })
            .option("--env-path", "Path to the .env file", {
            validator: program.STRING,
            global: true,
            default: 'env'
        })
            // information about the network
            .command("info", "Relevant information")
            .argument("<type>", "Type of information")
            .action(({ logger, args, options }) => __awaiter(this, void 0, void 0, function* () {
            const ctx = constants_1.contextEnv(options.envPath, options.network);
            if (args.type == 'addresses') {
                getAddressInfo(ctx, logger);
            }
            else if (args.type == 'balance') {
                yield getBalanceInfo(ctx, logger);
            }
            else if (args.type == 'network') {
                getNetworkInfo(ctx, logger);
            }
            else if (args.type == 'livenetwork') {
                // implement this nicely
            }
            else if (args.type == 'validators') {
                yield getValidatorInfo(ctx, logger);
            }
        }))
            // moving funds from one chain to another
            .command("crosschain", "Move funds from one chain to another")
            .argument("<type>", "Type of a crosschain transaction")
            .option("-a, --amount <amount>", "Amount to transfer", validatorstring)
            .option("-f, --fee <fee>", "Fee of a transaction", validatorstring)
            .action(({ logger, args, options }) => __awaiter(this, void 0, void 0, function* () {
            const ctx = constants_1.contextEnv(options.envPath, options.network);
            if (args.type == 'exportCP') {
                yield exportCP(ctx, logger, options.amount, options.fee);
            }
            else if (args.type == 'importCP') {
                yield importCP(ctx, logger);
            }
            else if (args.type == 'exportPC') {
                yield exportPC(ctx, logger, options.amount);
            }
            else if (args.type == 'importPC') {
                yield importPC(ctx, logger, options.fee);
            }
        }))
            // staking
            .command("stake", "Stake funds on the P-chain")
            .option("-n, --node-id <nodeID>", "The staking node's id", validatorstring)
            .option("-w, --weight <weight>", "Weight or amount to stake", validatorstring)
            .option("-d, --duration <duration>", "Duration of the staking process", validatorstring)
            .action(({ logger, args, options }) => __awaiter(this, void 0, void 0, function* () {
            const ctx = constants_1.contextEnv(options.envPath, options.network);
            yield stake(ctx, logger, options.nodeId, options.weight, options.duration);
        }))
            // hashing validator configuration
            .command("hash", "Utilities to calculate validator config hashes")
            .option("-n, --node-id <nodeID>", "The staking node's id", validatorstring)
            .option("-w, --weight <weight>", "Weight or amount to stake", validatorstring)
            .option("-d, --duration <duration>", "Duration of the staking process", validatorstring)
            .option("-a, --address <address>", "Validator's address in Bech32 format (default is derived from logged private key)", validatorstring)
            .action(({ logger, args, options }) => __awaiter(this, void 0, void 0, function* () {
            const ctx = constants_1.contextEnv(options.envPath, options.network);
            yield getHash(ctx, logger, options.nodeId, options.weight, options.duration, options.address);
        }))
            // converting between address formats
            .command("convert", "Utility for conversion of address formats")
            .argument("<type>", "Type of conversion")
            .option("-p, --public-key <pubk>", "User's secp256k1 public key", validatorstring)
            .action(({ logger, args, options }) => {
            const ctx = constants_1.contextEnv(options.envPath, options.network);
            if (args.type == 'PChainAddressFromPublicKey') {
                getAddressFromPublicKey(ctx, logger, options.publicKey);
            }
        });
    });
}
exports.cli = cli;
function getAddressInfo(ctx, logger) {
    const publicKey = utils_1.parseToID(ctx.pAddressBech32);
    logger.info(`X-chain address: ${ctx.xAddressBech32}`);
    logger.info(`P-chain address: ${ctx.pAddressBech32}`);
    logger.info(`C-chain address hex: ${ctx.cAddressHex}`);
    logger.info(`secp256k1 public key: ${publicKey}`);
}
function getBalanceInfo(ctx, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        let cbalance = (new dist_1.BN(yield ctx.web3.eth.getBalance(ctx.cAddressHex))).toString();
        let pbalance = (new dist_1.BN((yield ctx.pchain.getBalance(ctx.pAddressBech32)).balance)).toString();
        let xbalance = (new dist_1.BN((yield ctx.xchain.getBalance(ctx.xAddressBech32, ctx.avaxAssetID)).balance)).toString();
        cbalance = utils_1.integerToDecimal(cbalance, 18);
        pbalance = utils_1.integerToDecimal(pbalance, 9);
        xbalance = utils_1.integerToDecimal(xbalance, 9);
        logger.info(`${ctx.cAddressHex}: ${cbalance}`);
        logger.info(`${ctx.pAddressBech32}: ${pbalance}`);
        logger.info(`${ctx.xAddressBech32}: ${xbalance}`);
    });
}
function getNetworkInfo(ctx, logger) {
    const pchainId = ctx.pchain.getBlockchainID();
    const cchainId = ctx.cchain.getBlockchainID();
    const xchainId = ctx.xchain.getBlockchainID();
    logger.info(`blockchainId for P-chain: ${pchainId}`);
    logger.info(`blockchainId for C-chain: ${cchainId}`);
    logger.info(`blockchainId for X-chain: ${xchainId}`);
    logger.info(`assetId: ${ctx.avaxAssetID}`);
}
function getValidatorInfo(ctx, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        const pending = yield ctx.pchain.getPendingValidators();
        const current = yield ctx.pchain.getCurrentValidators();
        const fpending = JSON.stringify(pending, null, 2);
        const fcurrent = JSON.stringify(current, null, 2);
        logger.info(`pending: ${fpending}`);
        logger.info(`current: ${fcurrent}`);
    });
}
function exportCP(ctx, logger, amount, fee) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = new dist_1.BN(utils_1.decimalToInteger(amount, 9));
        const ffee = (fee === undefined) ?
            fee : new dist_1.BN(utils_1.decimalToInteger(fee, 9));
        const { txid, usedFee } = yield evmAtomicTx_1.exportTxCP(ctx, famount, ffee);
        if (fee !== usedFee)
            logger.info(`Used fee of ${usedFee}`);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function importCP(ctx, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        const { txid } = yield pvmAtomicTx_1.importTxCP(ctx);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function exportPC(ctx, logger, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        const famount = (amount === undefined) ?
            amount : new dist_1.BN(utils_1.decimalToInteger(amount, 9));
        const { txid } = yield pvmAtomicTx_1.exportTxPC(ctx, famount);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function importPC(ctx, logger, fee) {
    return __awaiter(this, void 0, void 0, function* () {
        const ffee = (fee === undefined) ?
            fee : new dist_1.BN(utils_1.decimalToInteger(fee, 9));
        const { txid, usedFee } = yield evmAtomicTx_1.importTxPC(ctx, ffee);
        if (fee !== usedFee)
            logger.info(`Used fee of ${usedFee}`);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function stake(ctx, logger, nodeID, weight, duration) {
    return __awaiter(this, void 0, void 0, function* () {
        const fweight = new dist_1.BN(utils_1.decimalToInteger(weight, 9));
        const fduration = new dist_1.BN(duration);
        const { txid } = yield addValidator_1.addValidator(ctx, nodeID, fweight, fduration);
        logger.info(`Success! TXID: ${txid}`);
    });
}
function getHash(ctx, logger, nodeID, weight, duration, address) {
    return __awaiter(this, void 0, void 0, function* () {
        if (address === undefined)
            address = ctx.pAddressBech32;
        const configHash = utils_1.toValidatorConfigHash(ctx.config.networkID.toString(), utils_1.parseToID(address), nodeID, utils_1.decimalToInteger(weight, 9), duration);
        logger.info(`Validator configuration hash: ${configHash}`);
    });
}
function getAddressFromPublicKey(ctx, logger, pubk) {
    const bech32 = utils_1.formatBech32(ctx.config.hrp, pubk);
    logger.info(`P-chain address: P-${bech32}`);
}
